from typing import Any, Dict, Union

import numpy as np
from scipy.sparse import coo_matrix, csr_matrix, vstack


def build_regularizer_matrix(
    data: Dict[str, Any],
    reg_type: str,
    smoothness: Union[float, np.ndarray]
) -> csr_matrix:
    """
    Constructs the specified regularizer matrix for gridfit.

    Parameters
    ----------
    data : Dict[str, Any]
        Dictionary containing necessary grid info, e.g.:
          - 'nx', 'ny', 'ngrid', 'dx', 'dy', 'xscale', 'yscale'
    reg_type : str
        Name of the regularizer: 'gradient', 'diffusion'/'laplacian', or 'springs'
    smoothness : float or np.ndarray
        Smoothing parameter (scalar or array)

    Returns
    -------
    csr_matrix
        The regularizer matrix of shape (m, ngrid), where m depends on the method.
    """
    if reg_type.lower() == "gradient":
        return _build_gradient_reg(data, smoothness)
    elif reg_type.lower() in ["diffusion", "laplacian"]:
        return _build_diffusion_reg(data, smoothness)
    elif reg_type.lower() == "springs":
        return _build_springs_reg(data, smoothness)
        # raise NotImplementedError("Springs regularizer is not implemented yet.")
    else:
        raise ValueError(f"Only 'gradient' regularizer is implemented now. Got {reg_type}")

def _build_gradient_reg(
    data: Dict[str, Any],
    smoothness: Union[float, np.ndarray]
) -> csr_matrix:
    """
    Builds the gradient-based regularizer (two PDE-like stencils),
    matching the MATLAB indexing logic for 'gradient'.

    Parameters
    ----------
    data : Dict[str, Any]
        - 'nx', 'ny': grid dimensions
        - 'ngrid': total number of grid points
        - 'dx', 'dy': 1D arrays of cell sizes
        - 'xscale', 'yscale': float scaling factors
    smoothness : float or np.ndarray
        Smoothing parameter.

    Returns
    -------
    csr_matrix
        The stacked gradient regularizer matrix.
    """

    nx = data["nx"]
    ny = data["ny"]
    ngrid = data["ngrid"]
    dx = data["dx"]  # length nx-1
    dy = data["dy"]  # length ny-1
    xscale = data["xscale"]
    yscale = data["yscale"]

    # Possibly handle anisotropic smoothing
    if np.isscalar(smoothness):
        xyRelative = np.array([1.0, 1.0])
    else:
        arr = np.asarray(smoothness, dtype=float)
        smoothparam = np.sqrt(np.prod(arr))
        xyRelative = arr / smoothparam

    # ---------------------------
    # 1) "y-gradient" portion
    #
    # MATLAB does:
    #   [i,j] = meshgrid(1:nx, 2:(ny-1));
    #   ind = j(:) + ny*(i(:)-1);
    #   dy1 = dy(j(:)-1)/yscale; etc.
    # ---------------------------
    i_vals = np.arange(1, nx+1)          # 1..nx
    j_vals = np.arange(2, ny)            # 2..(ny-1)
    j_grid, i_grid = np.meshgrid(j_vals, i_vals, indexing='xy')
    # j_grid.shape = (ny-2, nx)
    # i_grid.shape = (ny-2, nx)

    # Flatten
    i_flat = i_grid.ravel()
    j_flat = j_grid.ravel()

    # 1-based "ind" in MATLAB. 
    # Remember: j ranges [2..ny-1], i in [1..nx].
    # Then "ind = j(:) + ny*(i(:)-1)" in MATLAB
    ind_y = j_flat + ny*(i_flat - 1)  # still 1-based

    # dy1 = dy(j-1)/yscale; dy2 = dy(j)/yscale
    dy1 = dy[j_flat - 2]/yscale  # j=2 => j-2=0 => dy[0]
    dy2 = dy[j_flat - 1]/yscale  # j=2 => j-1=1 => dy[1]

    # The three coefficients for each row
    # corresponds to columns: (ind-1), (ind), (ind+1) in MATLAB
    # scaled by "xyRelative[1]" for the y dimension
    cvals_y = xyRelative[1] * np.column_stack([
        -2.0/(dy1*(dy1+dy2)), 
         2.0/(dy1*dy2),
        -2.0/(dy2*(dy1+dy2))
    ])

    # build row, col, data for the "y-grad" part
    row_y = np.repeat(ind_y, 3)  # each "ind" spawns 3 entries
    # col offsets are [ind-1, ind, ind+1] => still 1-based
    # but Python arrays are 0-based => subtract 1
    col_y = np.column_stack([
        ind_y - 1,
        ind_y,
        ind_y + 1
    ]).ravel() - 1  # 0-based
    data_y = cvals_y.ravel()

    y_grad_coo = coo_matrix((data_y, (row_y-1, col_y)), shape=(ngrid, ngrid))
    # NB: row_y-1 as well, so row indices are 0-based.

    # ---------------------------
    # 2) "x-gradient" portion
    #
    # MATLAB does:
    #   [i,j] = meshgrid(2:(nx-1), 1:ny);
    #   ind = j(:) + ny*(i(:)-1);
    #   dx1 = dx(i(:)-1)/xscale; etc.
    # ---------------------------
    i_vals = np.arange(2, nx)            # 2..(nx-1)
    j_vals = np.arange(1, ny+1)          # 1..ny
    j_grid, i_grid = np.meshgrid(j_vals, i_vals, indexing='xy')
    # j_grid.shape = (ny, nx-2)
    # i_grid.shape = (ny, nx-2)

    i_flat = i_grid.ravel()
    j_flat = j_grid.ravel()

    # same formula "ind = j + ny*(i-1)"
    ind_x = j_flat + ny*(i_flat - 1)  # 1-based

    dx1 = dx[i_flat - 2]/xscale  # i=2 => i-2=0 => dx[0]
    dx2 = dx[i_flat - 1]/xscale  # i=2 => i-1=1 => dx[1]

    cvals_x = xyRelative[0] * np.column_stack([
        -2.0/(dx1*(dx1+dx2)),
         2.0/(dx1*dx2),
        -2.0/(dx2*(dx1+dx2))
    ])

    row_x = np.repeat(ind_x, 3)
    col_x = np.column_stack([
        ind_x - ny,
        ind_x,
        ind_x + ny
    ]).ravel() - 1
    data_x = cvals_x.ravel()

    x_grad_coo = coo_matrix((data_x, (row_x-1, col_x)), shape=(ngrid, ngrid))

    # ---------------------------
    # 3) Stack them
    # ---------------------------
    Areg = vstack([y_grad_coo, x_grad_coo]).tocsr()

    # optionally remove all‐zero rows
    # this differs from MATLAB version, but is more efficient
    # row_sums = abs(Areg).sum(axis=1).A.ravel()
    # keep = (row_sums != 0)
    # Areg = Areg[keep]

    return Areg


def _build_diffusion_reg(
    data: Dict[str, Any],
    smoothness: Union[float, np.ndarray]
) -> csr_matrix:
    """
    Thermal diffusion or Laplacian regularizer, replicating
    the MATLAB code for {'diffusion','laplacian'}.

    Parameters
    ----------
    data : Dict[str, Any]
        - 'nx', 'ny', 'ngrid'
        - 'dx', 'dy' (1D arrays)
        - 'xscale', 'yscale'
    smoothness : float or np.ndarray
        Smoothing parameter.

    Returns
    -------
    csr_matrix
        The Laplacian regularizer matrix.
    """

    nx = data["nx"]
    ny = data["ny"]
    ngrid = data["ngrid"]
    dx = data["dx"]  # length = nx-1
    dy = data["dy"]  # length = ny-1
    xscale = data["xscale"]
    yscale = data["yscale"]

    # Handle anisotropy
    if np.isscalar(smoothness):
        xyRelativeStiffness = np.array([1., 1.])
    else:
        arr = np.asarray(smoothness, dtype=float)
        # We do the same "sqrt(prod(...))" logic if we want 
        # to replicate how MATLAB uses "smoothparam"
        # But for the ratio we only need arr / sqrt(prod(arr)).
        # Typically the user does that in the calling code anyway.
        xyRelativeStiffness = arr / np.sqrt(np.prod(arr))

    # ------------------------------------------------
    # y-direction Laplacian part
    # [i,j] = meshgrid(1:nx, 2:(ny-1));
    i_vals = np.arange(1, nx+1)     # 1..nx
    j_vals = np.arange(2, ny)       # 2..(ny-1)
    j_grid, i_grid = np.meshgrid(j_vals, i_vals, indexing='xy')
    i_flat = i_grid.ravel()
    j_flat = j_grid.ravel()

    # 1-based index
    ind_y = j_flat + ny*(i_flat - 1)

    # dy1, dy2
    dy1 = dy[j_flat - 2] / yscale
    dy2 = dy[j_flat - 1] / yscale

    cvals_y = xyRelativeStiffness[1] * np.column_stack([
        -2.0/(dy1*(dy1+dy2)),
         2.0/(dy1*dy2),
        -2.0/(dy2*(dy1+dy2))
    ])

    row_y = np.repeat(ind_y, 3)
    col_y = np.column_stack([ind_y - 1, ind_y, ind_y + 1]).ravel() - 1
    data_y = cvals_y.ravel()

    Areg_y = coo_matrix((data_y, (row_y - 1, col_y)), shape=(ngrid, ngrid))

    # ------------------------------------------------
    # x-direction Laplacian part
    # [i,j] = meshgrid(2:(nx-1), 1:ny);
    i_vals = np.arange(2, nx)      # 2..(nx-1)
    j_vals = np.arange(1, ny+1)    # 1..ny
    j_grid, i_grid = np.meshgrid(j_vals, i_vals, indexing='xy')
    i_flat = i_grid.ravel()
    j_flat = j_grid.ravel()

    ind_x = j_flat + ny*(i_flat - 1)

    dx1 = dx[i_flat - 2]/xscale
    dx2 = dx[i_flat - 1]/xscale

    cvals_x = xyRelativeStiffness[0] * np.column_stack([
        -2.0/(dx1*(dx1+dx2)),
         2.0/(dx1*dx2),
        -2.0/(dx2*(dx1+dx2))
    ])

    row_x = np.repeat(ind_x, 3)
    col_x = np.column_stack([ind_x - ny, ind_x, ind_x + ny]).ravel() - 1
    data_x = cvals_x.ravel()

    Areg_x = coo_matrix((data_x, (row_x - 1, col_x)), shape=(ngrid, ngrid))

    # Combine
    Areg = Areg_y + Areg_x  # sum of two coo_matrix => coo_matrix
    Areg = Areg.tocsr()

    return Areg



def _build_springs_reg(
    data: Dict[str, Any],
    smoothness: Union[float, np.ndarray]
) -> csr_matrix:
    """
    Spring-based regularizer: zero-rest-length springs along
    grid edges and diagonals, ported from the MATLAB implementation.

    Parameters
    ----------
    data : Dict[str, Any]
        - 'nx', 'ny', 'ngrid'
        - 'dx', 'dy'
        - 'xscale', 'yscale'
    smoothness : float or np.ndarray
        Smoothing parameter.

    Returns
    -------
    csr_matrix
        The "springs" regularizer matrix.
    """

    nx = data["nx"]
    ny = data["ny"]
    ngrid = data["ngrid"]
    dx = data["dx"]
    dy = data["dy"]
    xscale = data["xscale"]
    yscale = data["yscale"]

    if np.isscalar(smoothness):
        xyRelativeStiffness = np.array([1., 1.])
    else:
        arr = np.asarray(smoothness, dtype=float)
        xyRelativeStiffness = arr / np.sqrt(np.prod(arr))

    springs = []

    # -----------------------------------------
    # 1. Vertical springs 
    # [i,j] = meshgrid(1:nx,1:(ny-1));
    i_vals = np.arange(1, nx+1)     
    j_vals = np.arange(1, ny)
    j_grid, i_grid = np.meshgrid(j_vals, i_vals, indexing='xy')
    i_flat = i_grid.ravel()
    j_flat = j_grid.ravel()

    ind = j_flat + ny * (i_flat - 1)

    m = nx * (ny - 1)
    stiffness = 1.0 / (dy / yscale)
    vals = xyRelativeStiffness[1] * stiffness[j_flat - 1].ravel()[:, None] * np.array([-1, 1])
    row = np.repeat(np.arange(m), 2)
    col = np.column_stack([ind.ravel(), ind.ravel() + 1]).ravel() - 1
    data_vals = vals.ravel()
    Areg1 = coo_matrix((data_vals, (row, col)), shape=(m, ngrid))
    springs.append(Areg1)

    # -----------------------------------------
    # 2. Horizontal springs 
    # [i,j] = meshgrid(1:(nx-1),1:ny);
    i_vals = np.arange(1, nx)
    j_vals = np.arange(1, ny+1)      
    j_grid, i_grid = np.meshgrid(j_vals, i_vals, indexing='xy')  
    i_flat = i_grid.ravel()
    j_flat = j_grid.ravel()
    
    ind = j_flat + ny * (i_flat - 1)

    m = (nx - 1) * ny
    stiffness = 1.0 / (dx / xscale)
    vals = xyRelativeStiffness[0] * stiffness[i_flat - 1].ravel()[:, None] * np.array([-1, 1])
    row = np.repeat(np.arange(m), 2)
    col = np.column_stack([ind.ravel(), ind.ravel() + ny]).ravel() - 1
    data_vals = vals.ravel()
    Areg2 = coo_matrix((data_vals, (row, col)), shape=(m, ngrid))
    springs.append(Areg2)

    # -----------------------------------------
    # 3. Diagonal springs 
    # [i,j] = meshgrid(1:(nx-1),1:(ny-1));
    i_vals = np.arange(1, nx)
    j_vals = np.arange(1, ny)     
    j_grid, i_grid = np.meshgrid(j_vals, i_vals, indexing='xy')  
    i_flat = i_grid.ravel()
    j_flat = j_grid.ravel()
    
    ind = j_flat + ny * (i_flat - 1)

    m = (nx - 1) * (ny - 1)
    dx_scaled = dx[i_flat - 1] / xscale / xyRelativeStiffness[0]
    dy_scaled = dy[j_flat - 1] / yscale / xyRelativeStiffness[1]
    stiffness = 1.0 / np.sqrt(dx_scaled**2 + dy_scaled**2)
    vals = stiffness.ravel()[:, None] * np.array([-1, 1])
    row = np.repeat(np.arange(m), 2)
    col = np.column_stack([ind.ravel(), ind.ravel() + ny + 1]).ravel() - 1
    data_vals = vals.ravel()
    Areg3 = coo_matrix((data_vals, (row, col)), shape=(m, ngrid))
    springs.append(Areg3)

    # -----------------------------------------
    # 4. Diagonal springs (anti-diag: connects [i+1,j] to [i,j+1])
    col_shift = ny
    row_shift = 1
    col = np.column_stack([ind.ravel() + row_shift, ind.ravel() + col_shift])
    vals = stiffness.ravel()[:, None] * np.array([-1, 1])
    row = np.repeat(np.arange(m), 2)
    col = col.ravel() - 1
    data_vals = vals.ravel()
    Areg4 = coo_matrix((data_vals, (row, col)), shape=(m, ngrid))
    springs.append(Areg4)

    Areg = vstack(springs).tocsr()

    return Areg



import numpy as np
import pytest

from pygridfit.utils import check_params, validate_inputs


def test_check_params_happy_path():
    """
    Test a valid, representative set of parameters.
    """
    smoothness = [1.0, 2.0]
    extend = "warn"  # valid abbreviation for "warning"
    interpolation = "tri"  # valid abbreviation for "triangle"
    regularizer = "lapl"  # valid abbreviation for "laplacian" => becomes "diffusion"
    solver = "norm"  # => becomes "normal"
    tilesize = 100
    overlap = 0.3

    (
        final_smoothness,
        final_extend,
        final_interp,
        final_reg,
        final_solver,
        final_tilesize,
        final_overlap
    ) = check_params(
        smoothness,
        extend,
        interpolation,
        regularizer,
        solver,
        tilesize,
        overlap
    )

    # Assertions
    assert np.allclose(final_smoothness, np.array([1.0, 2.0]))
    assert final_extend == "warning"       # 'warn' -> 'warning'
    assert final_interp == "triangle"      # 'tri' -> 'triangle'
    assert final_reg == "diffusion"        # 'lapl' -> 'diffusion'
    assert final_solver == "normal"     # '\\' -> 'backslash'
    assert final_tilesize == 100
    assert final_overlap == 0.3

def test_check_params_smoothness_scalar():
    """
    Smoothness = scalar float > 0
    """
    (smooth, *_ ) = check_params(
        2.0, "never", "bilinear", "springs", "lsqr", None, None
    )
    assert smooth == 2.0

def test_check_params_smoothness_none_defaults_to_1():
    """
    If smoothness is None, it should default to 1.0
    """
    (smooth, *_ ) = check_params(
        None, "never", "bilinear", "springs", "lsqr", None, None
    )
    assert smooth == 1.0

@pytest.mark.parametrize("invalid_smoothness", [-1, 0])
def test_check_params_smoothness_invalid_scalar(invalid_smoothness):
    """
    Negative or zero scalar => 'Smoothness must be positive.'
    """
    with pytest.raises(ValueError, match="Smoothness must be positive"):
        check_params(
            smoothness=invalid_smoothness,
            extend="warning",
            interp="triangle",
            regularizer="gradient",
            solver="normal"
        )


@pytest.mark.parametrize("invalid_smoothness", [[1.0, -2], [1.0, 2.0, 3.0]])
def test_check_params_smoothness_invalid_array(invalid_smoothness):
    """
    Any negative or zero in the array, or length > 2 => 
    'Smoothness must be a positive scalar or 2-element vector.'
    """
    with pytest.raises(ValueError, match="Smoothness must be a positive scalar or 2-element vector"):
        check_params(
            smoothness=invalid_smoothness,
            extend="warning",
            interp="triangle",
            regularizer="gradient",
            solver="normal"
        )
 
def test_check_params_extend_invalid():
    """
    'foobar' is not a valid partial match => ValueError
    """
    with pytest.raises(ValueError, match="Invalid extend option: foobar"):
        check_params(1.0, "foobar", "triangle", "gradient", "normal")

@pytest.mark.parametrize("val", [2.9999, -10, 0])
def test_check_params_tilesize_invalid(val):
    """
    tilesize <3 => ValueError (unless inf)
    """
    with pytest.raises(ValueError, match="Tilesize must be >= 3 or inf"):
        check_params(1.0, "never", "bilinear", "diffusion", "normal", val, 0.2)

def test_check_params_tilesize_ok_inf():
    """
    tilesize=None => inf => pass
    """
    *_ , final_tilesize, _ = check_params(1.0, "never", "bilinear", "diffusion", "normal")
    assert final_tilesize == float("inf")

def test_check_params_overlap_default():
    """
    overlap=None => default 0.20
    """
    *_, final_tilesize, final_overlap = check_params(1.0, "never", "bilinear", "diffusion", "normal", 200, None)
    assert final_tilesize == 200
    assert final_overlap == 0.20

@pytest.mark.parametrize("bad_overlap", [-0.01, 0.51, 10])
def test_check_params_overlap_out_of_bounds(bad_overlap):
    with pytest.raises(ValueError, match="Overlap must be between 0 and 0.5"):
        check_params(1.0, "never", "triangle", "gradient", "normal", 100, bad_overlap)


# -----------------------------------------------------------------------------------
#                         validate_inputs tests
# -----------------------------------------------------------------------------------

def test_validate_inputs_happy_path():
    """
    Provide a valid set of x,y,z, plus integer xnodes,ynodes => check shape & defaults
    """
    x = np.linspace(0, 10, 50)
    y = np.linspace(5, 15, 50)
    z = np.linspace(-2, 2, 50)
    # Mark a few NaNs
    x[3] = np.nan
    y[10] = np.nan
    z[20] = np.nan

    result = validate_inputs(
        x, y, z,
        xnodes=5, ynodes=5,
        smoothness=1.5,
        maxiter=None,
        extend="warning",
        autoscale="on",
        xscale=1.0,
        yscale=1.0,
        interp="triangle",
        regularizer="gradient",
        solver="normal",
    )
    # Check that the 3 nans were removed
    # => final length is 47
    assert len(result["x"]) == 47
    assert len(result["y"]) == 47
    assert len(result["z"]) == 47

    assert result["nx"] == 5
    assert result["ny"] == 5
    assert result["ngrid"] == 25
    # autoscale => xscale, yscale are set to average spacing
    assert result["autoscale"] == "off"
    assert result["xscale"] == pytest.approx( (result["xnodes"][1:] - result["xnodes"][:-1]).mean() )
    assert result["yscale"] == pytest.approx( (result["ynodes"][1:] - result["ynodes"][:-1]).mean() )

    # Check maxiter => default = min(10000, ngrid)
    assert result["maxiter"] == 25

def test_validate_inputs_arrays_xnodes_ynodes():
    """
    If xnodes, ynodes are actual arrays, no changes except possibly extends if data is out of range
    """
    x = [1, 2, 3, 4, 5]
    y = [2, 4, 6, 8, 10]
    z = [-1, 0, 1, 2, 3]
    xnodes = np.array([0.0, 2.5, 5.0])
    ynodes = np.array([1.0, 5.5, 11.0])

    result = validate_inputs(
        x, y, z,
        xnodes=xnodes,
        ynodes=ynodes,
        smoothness=1.0,
        maxiter=10,
        extend="always",
        autoscale="off",
        xscale=1.0,
        yscale=1.0,
        interp="bilinear",
        regularizer="diffusion",
        solver="normal"
    )
    # Because extend="always", the min x < node_array[0], so node_array[0] => 1.0
    # Actually data’s min x=1, which is > original node_array[0]=0 => no extension needed on the low side
    # But data’s max x=5 => within node_array[-1]=5 => no extension needed
    # y data => min=2 => node_array[0]=1 => no extension needed
    # max=10 => node_array[-1]=11 => no extension needed

    # So we expect no changes. Let's verify
    assert np.allclose(result["xnodes"], xnodes)
    assert np.allclose(result["ynodes"], ynodes)
    assert result["dx"] == pytest.approx([2.5, 2.5])
    assert result["dy"] == pytest.approx([4.5, 5.5])
    assert result["nx"] == 3
    assert result["ny"] == 3
    assert result["ngrid"] == 9
    assert result["maxiter"] == 10
    assert result["autoscale"] == "off"
    assert result["xscale"] == 1.0

def test_validate_inputs_bad_nodes():
    """
    xnodes or ynodes non-monotonic => ValueError
    """
    x = np.linspace(0, 5, 10)
    y = np.linspace(0, 5, 10)
    z = np.linspace(-1, 1, 10)
    xnodes = [0, 3, 2]  # not strictly increasing
    with pytest.raises(ValueError, match="must be strictly increasing"):
        validate_inputs(
            x, y, z,
            xnodes=xnodes,
            ynodes=3,
            smoothness=1.0,
            maxiter= None,
            extend="warning",
            autoscale="off",
            xscale=1.0,
            yscale=1.0,
            interp="triangle",
            regularizer="gradient",
            solver="normal",
        )

def test_validate_inputs_insufficient_points():
    """
    if after removing NaNs, fewer than 3 points => ValueError
    """
    x = [1, 2, np.nan]
    y = [5, 6, np.nan]
    z = [10, 11, np.nan]
    with pytest.raises(ValueError, match="Insufficient data"):
        validate_inputs(
            x, y, z,
            xnodes=3, ynodes=3,
            smoothness=1.0,
            maxiter=10,
            extend="warning",
            autoscale="off",
            xscale=1.0,
            yscale=1.0,
            interp="bilinear",
            regularizer="gradient",
            solver="normal",
        )



import numpy as np
import scipy.io

from pygridfit import GridFit

# Test against results from MATALB
# input = scipy.io.loadmat("./tests/data/input.mat")
# output = scipy.io.loadmat("./tests/data/output.mat")
bluff_data = scipy.io.loadmat("./tests/data/bluff_data.mat")


def test_gridfit():
    # Test the gridfit function
    x = bluff_data["x"].flatten()
    y = bluff_data["y"].flatten()
    z = bluff_data["z"].flatten()
    gx = np.arange(0, 268, 4)
    gy = np.arange(0, 404, 4)

    gf = GridFit(
        x, y, z,
        xnodes=gx,
        ynodes=gy,
        smoothness=1,
        interp="triangle",
        regularizer="gradient",
        solver="normal",
    )
    gf.fit()
    
    assert np.allclose(gf.zgrid, bluff_data["g"])


from typing import Any, Dict, Optional, Union, cast

import numpy as np
from numpy.typing import NDArray


def _resolve_abbrev(value: str, valid_options: list[str], fieldname: str) -> str:
    """
    Resolve a possibly-abbreviated string 'value' to one of the entries in valid_options,
    if there is exactly one match. If none or more than one match, raise ValueError.

    Parameters
    ----------
    value : str
        The input string to match.
    valid_options : list[str]
        A list of valid strings (full options).
    fieldname : str
        A descriptive name of the field for error messages.

    Returns
    -------
    str
        The matched (fully expanded) option from valid_options.
    """
    val_lower = value.lower()
    matches = [opt for opt in valid_options if opt.startswith(val_lower)]
    if len(matches) == 1:
        return matches[0]
    elif len(matches) == 0:
        raise ValueError(f"Invalid {fieldname} option: {value}")
    else:
        raise ValueError(f"Ambiguous {fieldname} option: {value}")



def check_params(
    smoothness: Union[float, list[float], NDArray[np.float64], None],
    extend: str,
    interp: str,
    regularizer: str,
    solver: str,
    tilesize: Optional[float] = None,
    overlap: Optional[float] = None,
) -> tuple[Union[NDArray[np.float64], float], str, str, str, str, float, float]:
    """
    Validate and standardize the gridfit parameters, mimicking the MATLAB check_params logic.

    Parameters
    ----------
    smoothness : float | list[float] | NDArray[np.float64] | None
        Desired smoothing parameter (scalar or up to 2-element array).
    extend : str
        Behavior for data outside node bounds. One of ["never","warning","always"].
    interp : str
        Interpolation type. One of ["bilinear","nearest","triangle"].
    regularizer : str
        Regularizer type. One of ["springs","diffusion","laplacian","gradient"].
    solver : str
        Solver type. One of ["symmlq","lsqr","normal"].
    tilesize : float, optional
        Size of each tile if using tiling, otherwise inf to disable tiling.
    overlap : float, optional
        Overlap fraction (0 <= overlap <= 0.5) between tiles if tiling is used.

    Returns
    -------
    tuple
        A 7-element tuple:
        (smoothness, extend, interp, regularizer, solver, tilesize, overlap),
        where smoothness is either a float or an array, and the rest are
        strings/floats in validated form.
    """

    # ----------------------------------------------------------------
    # 1) Validate smoothness
    # ----------------------------------------------------------------
    if smoothness is None:
        smoothness = 1.0
    elif isinstance(smoothness, (list, tuple, np.ndarray)):
        smoothness = np.array(smoothness, dtype=float).flatten()
        if smoothness.size > 2 or np.any(smoothness <= 0):
            raise ValueError("Smoothness must be a positive scalar or 2-element vector.")
    else:
        # Must be scalar > 0 if it’s just a float
        if smoothness <= 0:
            raise ValueError("Smoothness must be positive.")

    # ----------------------------------------------------------------
    # 2) Validate extend
    # ----------------------------------------------------------------
    valid_extend = ["never", "warning", "always"]
    extend = _resolve_abbrev(extend, valid_extend, "extend")

    # ----------------------------------------------------------------
    # 3) Validate interpolation
    #    (MATLAB: 'bilinear','nearest','triangle')
    # ----------------------------------------------------------------
    valid_interp = ["bilinear", "nearest", "triangle"]
    interp = _resolve_abbrev(interp, valid_interp, "interp")

    # ----------------------------------------------------------------
    # 4) Validate regularizer
    #    (MATLAB: 'springs','diffusion','laplacian','gradient')
    #    note that 'diffusion' and 'laplacian' are synonyms.
    # ----------------------------------------------------------------
    valid_reg = ["springs", "diffusion", "laplacian", "gradient"]
    regularizer = _resolve_abbrev(regularizer, valid_reg, "regularizer")
    # unify 'laplacian' into 'diffusion'
    if regularizer == "laplacian":
        regularizer = "diffusion"

    # ----------------------------------------------------------------
    # 5) Validate solver
    #    (MATLAB: 'symmlq','lsqr','normal')
    # ----------------------------------------------------------------
    valid_solver = ["symmlq", "lsqr", "normal"]
    solver = _resolve_abbrev(solver, valid_solver, "solver")

    # ----------------------------------------------------------------
    # 6) Validate tilesize
    # ----------------------------------------------------------------
    if tilesize is None:
        tilesize = float("inf")
    else:
        if tilesize < 3 and tilesize != float("inf"):
            raise ValueError("Tilesize must be >= 3 or inf (to disable tiling).")

    # ----------------------------------------------------------------
    # 7) Validate overlap
    # ----------------------------------------------------------------
    if overlap is None:
        overlap = 0.20
    else:
        if overlap < 0 or overlap > 0.5:
            raise ValueError("Overlap must be between 0 and 0.5 (inclusive).")

    return (smoothness, extend, interp, regularizer, solver, tilesize, overlap)
  

def validate_inputs(
    x: Union[NDArray[np.float64], list[float]],
    y: Union[NDArray[np.float64], list[float]],
    z: Union[NDArray[np.float64], list[float]],
    xnodes: Union[NDArray[np.float64], int],
    ynodes: Union[NDArray[np.float64], int],
    smoothness: Union[float, NDArray[np.float64]],
    maxiter: Optional[int],
    extend: str,
    autoscale: str,
    xscale: float,
    yscale: float,
    interp: str,
    regularizer: str,
    solver: str,
) -> Dict[str, Any]:
    """
    Preprocess and validate inputs in a style similar to the beginning of gridfit.m.
    Returns a dictionary of 'prepared' data needed by the solver or next step.

    Parameters
    ----------
    x, y, z : array-like of float
        Data coordinates and values. Each must be the same length.
    xnodes, ynodes : array-like of float or int
        If int, we auto-generate that many nodes from min->max of x or y.
        If array-like, must be strictly increasing.
    smoothness : float | NDArray[np.float64]
        Smoothing parameter, typically validated earlier.
    maxiter : int or None
        Iteration limit for iterative solvers; if None, a default is chosen.
    extend : str
        Bound extension behavior, e.g. 'never','warning','always'.
    autoscale : str
        If 'on', xscale, yscale are set to mean cell size. Then turned 'off'.
    xscale, yscale : float
        Scaling parameters used for regularization weight (set if autoscale='on').
    interp : str
        Interpolation method, e.g. 'bilinear','nearest','triangle'.
    regularizer : str
        The method for building a regularizer, e.g. 'diffusion','gradient','springs'.
    solver : str
        One of 'symmlq','lsqr','normal'.

    Returns
    -------
    Dict[str, Any]
        A dictionary containing validated and computed fields:
        {
          "x", "y", "z",
          "xnodes", "ynodes", "dx", "dy", "nx", "ny", "ngrid",
          "xmin", "xmax", "ymin", "ymax",
          "smoothness", "maxiter", "extend", "autoscale",
          "xscale", "yscale", "interp", "regularizer", "solver",
          "ind", "indx", "indy", "tx", "ty", ...
        }
    """

    # Convert x, y, z to flat numpy arrays
    x = np.asarray(x, dtype=float).ravel()
    y = np.asarray(y, dtype=float).ravel()
    z = np.asarray(z, dtype=float).ravel()

    # Remove NaNs
    nan_mask = (~np.isnan(x)) & (~np.isnan(y)) & (~np.isnan(z))
    x, y, z = x[nan_mask], y[nan_mask], z[nan_mask]

    if len(x) < 3:
        raise ValueError("Insufficient data for surface estimation (need at least 3 non-NaN points).")

    xmin, xmax = float(x.min()), float(x.max())
    ymin, ymax = float(y.min()), float(y.max())

    # Expand xnodes, ynodes if scalar
    if np.isscalar(xnodes):
        xnodes_arr = np.linspace(xmin, xmax, cast(int, xnodes))
        # Force the final node to match the data max, as in MATLAB
        xnodes_arr[-1] = xmax
    else:
        xnodes_arr = np.asarray(xnodes, dtype=float).ravel()

    if np.isscalar(ynodes):
        ynodes_arr = np.linspace(ymin, ymax, cast(int, ynodes))
        ynodes_arr[-1] = ymax
    else:
        ynodes_arr = np.asarray(ynodes, dtype=float).ravel()

    # Check for strictly increasing nodes
    dx = np.diff(xnodes_arr)
    dy = np.diff(ynodes_arr)
    if np.any(dx <= 0) or np.any(dy <= 0):
        raise ValueError("xnodes and ynodes must be strictly increasing.")

    nx, ny = len(xnodes_arr), len(ynodes_arr)
    ngrid = nx * ny

    # If autoscale is 'on', set xscale, yscale internally
    if autoscale.lower() == "on":
        xscale = float(dx.mean())
        yscale = float(dy.mean())
        autoscale = "off"  # turn off after applying once

    # If maxiter is not specified, pick a default
    if maxiter is None or maxiter == "":
        maxiter = min(10000, ngrid)

    # Check x, y, z lengths
    if len(x) != len(y) or len(x) != len(z):
        raise ValueError("x, y, z must be of the same length.")

    # Function to adjust node arrays if data extends beyond them
    def maybe_extend(
        bound_val: float,
        node_array: NDArray[np.float64],
        side: str,
        axis: str
    ) -> None:     
        """
        Possibly extend the node_array (in-place) if 'extend' is not 'never'.
        side is 'start' or 'end' to indicate which boundary we are checking.
        axis is 'x' or 'y' for error messages.
        """   
        if side == "start":
            if bound_val < node_array[0]:
                if extend == "always":
                    node_array[0] = bound_val
                elif extend == "warning":
                    print(
                        f"[GRIDFIT:extend] {axis}nodes(1) was decreased by "
                        f"{node_array[0] - bound_val:.6f}, new = {bound_val:.6f}"
                    )
                    node_array[0] = bound_val
                elif extend == "never":
                    raise ValueError(
                        f"Some {axis} ({bound_val}) falls below {axis}nodes(1) by {node_array[0] - bound_val:.6f}"
                    )
        elif side == "end":
            if bound_val > node_array[-1]:
                if extend == "always":
                    node_array[-1] = bound_val
                elif extend == "warning":
                    print(
                        f"[GRIDFIT:extend] {axis}nodes(end) was increased by "
                        f"{bound_val - node_array[-1]:.6f}, new = {bound_val:.6f}"
                    )
                    node_array[-1] = bound_val
                elif extend == "never":
                    raise ValueError(
                        f"Some {axis} ({bound_val}) falls above {axis}nodes(end) by {bound_val - node_array[-1]:.6f}"
                    )

    # Possibly extend boundaries
    maybe_extend(xmin, xnodes_arr.astype(np.float64), "start", "x")
    maybe_extend(xmax, xnodes_arr.astype(np.float64), "end", "x")
    maybe_extend(ymin, ynodes_arr.astype(np.float64), "start", "y")
    maybe_extend(ymax, ynodes_arr.astype(np.float64), "end", "y")

    # Recompute dx, dy because we may have changed xnodes/ynodes
    dx = np.diff(xnodes_arr)
    dy = np.diff(ynodes_arr)

    indx = np.digitize(x, xnodes_arr)
    indy = np.digitize(y, ynodes_arr)

    indx[indx == nx] -= 1
    indy[indy == ny] -= 1

    ind = indy + ny * (indx - 1)

    # Compute interpolation weights (tx, ty)
    tx = np.clip((x - xnodes_arr[indx - 1]) / dx[indx - 1], 0, 1)
    ty = np.clip((y - ynodes_arr[indy - 1]) / dy[indy - 1], 0, 1)

    # Just return everything. 
    return {
        "x": x,
        "y": y,
        "z": z,
        "xnodes": xnodes_arr,
        "ynodes": ynodes_arr,
        "dx": dx,
        "dy": dy,
        "nx": nx,
        "ny": ny,
        "ngrid": ngrid,
        "xmin": xmin,
        "xmax": xmax,
        "ymin": ymin,
        "ymax": ymax,
        "smoothness": smoothness,
        "maxiter": maxiter,
        "extend": extend,
        "autoscale": autoscale,
        "xscale": xscale,
        "yscale": yscale,
        "interp": interp,
        "regularizer": regularizer,
        "solver": solver,
        "ind": ind,
        "indx": indx,
        "indy": indy,
        "tx": tx,
        "ty": ty,
    }
